# encoding: utf-8
import functools
from tornado.ioloop import IOLoop
from tornado.gen import coroutine, Return


class AsyncCursor(object):

    def __init__(self, connection, thread_pool, waiter, on_open=None, on_close=None, **kwargs):
        self.__cursor = connection.cursor(**kwargs)
        self.__thread_pool = thread_pool
        self.__on_open = on_open
        self.__on_close = on_close
        self.__io_loop = IOLoop.current()
        self.__wait = waiter

        if callable(on_open):
            on_open(self)

        for prop in ('fetchall', 'fetchone', 'execute', 'fetchmany', 'callproc'):
            setattr(self, prop, self.__futurize(
                getattr(self.__cursor, prop),
            ))

    def __futurize(self, func):
        @functools.wraps(func)
        @coroutine
        def wrap(*args, **kwargs):
            yield self.__wait()
            result = yield self.__thread_pool.submit(func, *args, **kwargs)
            raise Return(result)
        return wrap

    def close(self):
        @coroutine
        def closer():
            yield self.__wait()
            self.__cursor.close()
            if callable(self.__on_close):
                self.__on_close(self)

        self.__io_loop.add_callback(closer)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def __getattr__(self, item):
        return getattr(self.__cursor, item)



#!/usr/bin/env python
# encoding: utf-8
import functools
import logging
from psycopg2._psycopg import QueryCanceledError
from concurrent import futures
from psycopg2 import connect
import psycopg2

from tornado.ioloop import IOLoop
from tornado.gen import Future, coroutine, sleep, Return
from tornado.process import cpu_count
from tornado.queues import Queue
from tornado_psycopg2.cursor import AsyncCursor

log = logging.getLogger("psycopg.async.connection")


class AsyncConnection(object):

    def __init__(self, *args, **kwargs):
        kwargs['async'] = True

        if "thread_pool" in kwargs:
            self.__thread_pool = kwargs.pop('thread_pool')
        else:
            self.__thread_pool = futures.ThreadPoolExecutor(cpu_count())

        self.__connection = connect(*args, **kwargs)

        self.__io_loop = IOLoop.current()
        self.__connected = False

        log.debug("Trying to connect to postgresql")
        f = self.__wait()
        self.__io_loop.add_future(f, self.__on_connect)
        self.__queue = Queue()
        self.__has_active_cursor = False

        for method in ('get_backend_pid', 'get_parameter_status'):
            setattr(self, method, self.__futurize(method))

    def __on_connect(self, result):
        log.debug("Connection establishment")
        self.__connected = True
        self.__io_loop.add_callback(self._loop)

    @coroutine
    def _loop(self):
        log.debug("Starting queue loop")
        while self.__connected:
            while self.__has_active_cursor or self.__connection.isexecuting():
                yield sleep(0.001)

            func, future = yield self.__queue.get()
            result = func()
            if isinstance(result, Future):
                result = yield result

            self.__io_loop.add_callback(future.set_result, result)
            yield self.__wait()

    @coroutine
    def __wait(self):
        log.debug("Waiting for events")
        while not (yield sleep(0.001)):
            try:
                state = self.__connection.poll()
            except QueryCanceledError:
                yield sleep(0.1)
                continue

            f = Future()

            def resolve(fileno, io_op):
                if f.running():
                    f.set_result(True)
                self.__io_loop.remove_handler(fileno)

            if state == psycopg2.extensions.POLL_OK:
                raise Return(True)

            elif state == psycopg2.extensions.POLL_READ:
                self.__io_loop.add_handler(self.__connection.fileno(), resolve, IOLoop.READ)
                yield f

            elif state == psycopg2.extensions.POLL_WRITE:
                self.__io_loop.add_handler(self.__connection.fileno(), resolve, IOLoop.WRITE)
                yield f

    def __on_cursor_open(self, cursor):
        self.__has_active_cursor = True
        log.debug('Opening cursor')

    def __on_cursor_close(self, cursor):
        self.__has_active_cursor = False
        log.debug('Closing active cursor')

    def cursor(self, **kwargs):
        f = Future()
        self.__io_loop.add_callback(
            self.__queue.put,
            (functools.partial(
                AsyncCursor,
                self.__connection,
                self.__thread_pool,
                self.__wait,
                on_open=self.__on_cursor_open,
                on_close=self.__on_cursor_close,
                **kwargs
            ), f)
        )
        return f

    def cancel(self):
        return self.__thread_pool.submit(self.__connection.cancel)

    def close(self):
        self.__has_active_cursor = True

        @coroutine
        def closer():
            while not (yield self.__queue.empty()):
                func, future = yield self.__queue.get()
                future.set_exception(psycopg2.Error("Connection closed"))

            self.__io_loop.add_callback(self.__connection.close)

    def __futurize(self, item):
        attr = getattr(self.__connection, item)

        @functools.wraps(attr)
        def wrap(*args, **kwargs):
            f = Future()
            self.__io_loop.add_callback(
                self.__queue.put,
                (functools.partial(attr, *args, **kwargs), f)
            )
            return f
        return wrap



#!/usr/bin/env python
# encoding: utf-8
import logging
from random import choice, randint
from tornado.gen import coroutine
from tornado.testing import gen_test
from psycopg2.extras import NamedTupleCursor
from test import TestBase


log = logging.getLogger("tests")


class ConnectionTestCase(TestBase):

    @gen_test
    def test_001_cursor(self):
        cursor = yield self.connection.cursor()
        yield cursor.execute("SELECT 1 as test")
        r = yield cursor.fetchone()
        self.assertEqual(r, (1,))
        cursor.close()

    @gen_test
    def test_002_cancel(self):
        for _ in range(randint(1, 20)):
            with (yield self.connection.cursor()) as cursor:
                yield cursor.execute('SELECT pg_sleep(60)')
                yield self.connection.cancel()

    @gen_test
    def test_003_multiple_cursors(self):
        for x in range(randint(100, 200)):
            with (yield self.connection.cursor()) as cursor:
                yield cursor.execute("SELECT %s as test" % x)
                r = yield cursor.fetchone()
                self.assertEqual(r, (x,))

    @gen_test
    def test_004_parallel_multiple_cursors(self):
        connection = self.connection

        @coroutine
        def test1():
            with (yield connection.cursor()) as cursor:
                yield cursor.execute("SELECT 100 as test")
                self.assertEqual((yield cursor.fetchone()), (100,))

        @coroutine
        def test2():
            with (yield connection.cursor()) as cursor:
                yield cursor.execute("SELECT 200 as test")
                self.assertEqual((yield cursor.fetchone()), (200,))

        yield [choice((test1, test2))() for _ in range(randint(100, 200))]

    @gen_test
    def test_005_get_backend_pid(self):
        self.assertTrue((yield self.connection.get_backend_pid()) > 0)

    @gen_test
    def test_006_get_parameter_status(self):
        result = yield self.connection.get_parameter_status('TIME ZONE')
        self.assertTrue(result is None)

    @gen_test
    def test_007_namedtuple_cursor(self):
        with (yield self.connection.cursor(cursor_factory=NamedTupleCursor)) as cursor:
            yield cursor.execute("select 1 as result")
            result = yield cursor.fetchone()

        self.assertEqual(result.result, 1)



#!/usr/bin/env python
# encoding: utf-8
import re
from setuptools import setup
from tornado_psycopg2 import __author__, __version__

author = re.compile(r"(?P<name>[^\<]+)\s{1,}\<(?P<email>.*)\>").match(__author__).groupdict()

setup(
    name='tornado-psycopg2',
    version=__version__,
    author=author['name'],
    author_email=author['email'],
    license="MIT",
    platforms="all",
    classifiers=(
        'Environment :: Console',
        'Programming Language :: Python',
    ),
    long_description=open('README.rst').read(),
    description='Tornado driver for support asynchronous mode for psycopg2.',
    include_package_data=False,
    packages=['tornado_psycopg2'],
    requires=['Python (>2.6)'],
    install_requires=[i.strip() for i in open('requirements.txt').read().split('\n') if i],
)
